"""Logger for Awstaga."""
from conflog import Conflog

def init(dry_run: bool):
    """Initialize logger.
    """

    labels = ''
    if dry_run:
        labels = '[dry-run] '

    cfl = Conflog(
        conf_dict={
            'level': 'info',
            'format': f'{labels}[awstaga] %(levelname)s %(message)s'
        }
    )

    return cfl.get_logger(__name__)



"""This module contains the Resource class."""

class Resource():
    """This class represents the resource to be tagged.
    It contains the ARN of the resource, the tags to be applied to the resource,
    and the names of the TagSets to be applied to the resource.
    """

    def __init__(self, arn: str, tags: list, tagset_names: list) -> None:
        """Initialize the Resource object."""
        self.arn = arn
        self.tags = tags
        self.tagsetnames = tagset_names

    def get_arn(self) -> str:
        """Return the ARN of the resource."""
        return self.arn

    def get_tags(self) -> list:
        """Return the tags that should be applied to this resource."""
        return self.tags

    def get_tagset_names(self) -> list:
        """Return the names of the tag sets which contain the tags that
        should be applied to this resource."""
        return self.tagsetnames



"""This module contains the Tag class."""

class Tag():
    """This class represents a tag.
    It contains the key and value pair of the tag.
    """

    def __init__(self, key: str, value: str) -> None:
        """Initialize the Tag object."""
        self.key = key
        self.value = value

    def get_key(self) -> str:
        """Return the key of the tag."""
        return self.key

    def get_value(self) -> str:
        """Return the value of the tag."""
        return self.value

    def __str__(self) -> str:
        """Return the string representation of the Tag object."""
        return f'{self.key}={self.value}'

    def __repr__(self) -> str:
        """Return the string representation of the Tag object."""
        return f'{self.key}={self.value}'



# pylint: disable=too-many-locals
"""
awstaga
=======
Tag any AWS resource via config file.
"""

import time
import json
import boto3
import click
from .config import load
from .logger import init

def apply(conf_file: str, dry_run: bool, batch_size: int, delay: int) -> None:
    """Apply tags to resources based on configuration file."""

    logger = init(dry_run)

    logger.info(f'Loading configuration file {conf_file}')
    conf = load(conf_file, dry_run)
    tagsets = conf[0]
    resources = conf[1]

    client = boto3.client('resourcegroupstaggingapi')

    batches = {}

    for resource in resources:
        tags = {}

        for tagset_name in resource.get_tagset_names():
            tagset = tagsets[tagset_name]
            for tagset_tag in tagset.get_tags():
                tags[tagset_tag.get_key()] = tagset_tag.get_value()

        for resource_tag in resource.get_tags():
            tags[resource_tag.get_key()] = resource_tag.get_value()

        if not tags:
            logger.warning(f'No tags to apply to resource {resource.get_arn()}')
        else:
            # Create a batch of resources with the same tags
            # the batch ID is derived from the tags dump.
            # Tags are purposely not sorted in order to distinguish
            # varying tag orders as defined by the users.
            batch_id = json.dumps(tags)
            if batch_id not in batches:
                batches[batch_id] = {'resource_arns': [], 'tags': tags}
            batches[batch_id]['resource_arns'].append(resource.get_arn())

            logger.info(f'Adding resource {resource.get_arn()} to a batch with tags {tags}')

            # Apply tags to current batch if batch_size is reached,
            # and the batch is then removed.
            if len(batches[batch_id]['resource_arns']) == batch_size:
                _process_batch(dry_run, delay, logger, client, batches[batch_id])
                batches.pop(batch_id)

    # Apply tags to remaining batches, each batch would have less than batch_size resources.
    for batch_id, batch in batches.items():
        _process_batch(dry_run, delay, logger, client, batch)

def _process_batch(dry_run: bool, delay: int, logger, client, batch: dict) -> None:
    if dry_run is False:
        logger.info(f'Applying {len(batch["resource_arns"])} resource(s) '\
                    f'with tags {batch["tags"]}')
        response = client.tag_resources(
                ResourceARNList=batch['resource_arns'],
                Tags=batch['tags']
            )
        if 'FailedResourcesMap' in response and len(response["FailedResourcesMap"].keys()) > 0:
            logger.error('Failed to apply tags to '\
                         f'{len(response["FailedResourcesMap"].keys())} resource(s):')
            for resource_arn, error in response['FailedResourcesMap'].items():
                logger.error(f'{resource_arn}: '\
                             f'{error.get("StatusCode")} - '\
                             f'{error.get("ErrorCode")} - '\
                             f'{error.get("ErrorMessage")}')
        time.sleep(delay)

@click.command()
@click.option('--conf-file', default='awstaga.yaml', show_default=True, type=str,
              help='Configuration file path')
@click.option('--dry-run', is_flag=True, default=False, show_default=True, type=bool,
              help='When dry run is enabled, no tags are applied')
@click.option('--batch-size', default=5, show_default=True, type=int,
              help='Number of resources to tag in one go per batch')
@click.option('--delay', default=2, show_default=True, type=int,
              help='Delay in seconds after tagging each batch')
def cli(conf_file: str, dry_run: bool, batch_size: int, delay: int) -> None:
    """Python CLI for tagging AWS resources based on a YAML configuration.
    """
    apply(conf_file, dry_run, batch_size, delay)



"""Configuration loader.
"""

from typing import Tuple
import os
import yaml
from yamlinclude import YamlIncludeConstructor
from .models.resource import Resource
from .models.tag import Tag
from .models.tagset import TagSet
from .logger import init

def load(conf_file: str, dry_run: bool) -> Tuple[dict, list]:
    """Load configuration values from file.
    """

    logger = init(dry_run)
    YamlIncludeConstructor.add_to_loader_class(
        loader_class=yaml.FullLoader,
        base_dir=os.path.dirname(conf_file)
    )

    tagsets = {}
    resources = []
    with open(conf_file, 'r', encoding='utf-8') as stream:
        conf_yaml = yaml.load(stream, Loader=yaml.FullLoader)
        if conf_yaml is not None:
            for key, value in conf_yaml.items():

                if key == 'tagsets':
                    if value is not None:
                        logger.info(f'Loading {len(value)} tagset(s)...')
                        _load_tagsets(logger, tagsets, value)
                    else:
                        logger.warning('No tagsets found in configuration file')

                elif key == 'resources':
                    if value is not None:
                        logger.info(f'Loading {len(value)} resource(s)...')
                        _load_resources(logger, resources, value)
                    else:
                        logger.warning('No resources found in configuration file')

    return tagsets, resources

def _load_tagsets(logger, tagsets: dict, tagsets_conf: list) -> None:
    for tagset in tagsets_conf:
        name = tagset['name']
        tags = []
        for tag in tagset['tags']:
            tags.append(Tag(tag['key'], tag['value']))
        logger.debug(f'Loaded tagset {name} with tags {*tags,}')
        tagsets[name] = TagSet(name, tags)

def _load_resources(logger, resources: list, resources_conf: list) -> None:
    for resource in resources_conf:
        arn = resource['arn']
        tags = []
        if 'tags' in resource:
            for tag in resource['tags']:
                tags.append(Tag(tag['key'], tag['value']))
        tagset_names = []
        if 'tagsetnames' in resource:
            for tagsetname in resource['tagsetnames']:
                tagset_names.append(tagsetname)
        if tags and tagset_names:
            logger.debug(f'Loaded resource {arn} with '\
                        f'tags {*tags,} and tagsetnames {*tagset_names,}')
        elif tags and not tagset_names:
            logger.debug(f'Loaded resource {arn} with '\
                        f'tags {*tags,}')
        elif not tags and tagset_names:
            logger.debug(f'Loaded resource {arn} with '\
                        f'tagsetnames {*tagset_names,}')
        resources.append(Resource(arn, tags, tagset_names))



"""This module contains the TagSet class."""

class TagSet():
    """This class represents a TagSet.
    It contains the name of the TagSet and the tags that belong to .
    """

    def __init__(self, name: str, tags: list) -> None:
        """Initialize the TagSet object."""
        self.name = name
        self.tags = tags

    def get_name(self) -> str:
        """Return the name of the tag set."""
        return self.name

    def get_tags(self) -> list:
        """Return the tags on this tag set."""
        return self.tags
